"""
This module defines an UFL transformation, that takes a UFL expression
and transforms it into a list of accumulation terms.
"""
from dune.perftool.ufl.flatoperators import construct_binary_operator
from dune.perftool.ufl.modified_terminals import extract_modified_arguments
from dune.perftool.ufl.transformations import ufl_transformation
from dune.perftool.ufl.transformations.replace import replace_expression
from dune.perftool.ufl.transformations.identitypropagation import identity_propagation
from dune.perftool.ufl.transformations.zeropropagation import zero_propagation
from dune.perftool.ufl.transformations.reindexing import reindexing
from dune.perftool.ufl.modified_terminals import analyse_modified_argument, ModifiedArgument
from dune.perftool.pdelab.restriction import Restriction

from ufl.classes import Zero, Identity, Indexed, IntValue, MultiIndex, Product
from ufl.core.multiindex import indices

from pytools import Record


class AccumulationTerm(Record):
    def __init__(self,
                 term,
                 argument,
                 indexmap={},
                 new_indices=None
                 ):
        assert isinstance(argument, ModifiedArgument)
        Record.__init__(self,
                        term=term,
                        argument=argument,
                        indexmap=indexmap,
                        new_indices=new_indices
                        )


@ufl_transformation(name="accterms", extraction_lambda=lambda l: [at.term for at in l])
def split_into_accumulation_terms(expr):
    """Split an UFL expression into several accumulation parts and return a list

    For a residual evaluation we split for different test functions
    and according to the restriction (sefl/neighbor at skeletons). For
    the jacobians we also need to split according to the ansatz
    functions (and their restriction).

    Note: This function is not an UFL transformation. Nonetheless it
    has the @ufl_transformation decorator for debugging purposes.

    Arguments:
    ----------
    expr: UFL expression we want to split
    """
    # Store AccumulationTerms in this list
    ret = []

    # Extract a list of modified terminals for the test function
    # One accumulation instruction will be generated for each of these.
    test_args = extract_modified_arguments(expr, argnumber=0, do_index=False)

    # Extract a list of modified terminals for the ansatz function
    # in jacobian forms.
    all_jacobian_args = extract_modified_arguments(expr, argnumber=1, do_index=False)

    for test_arg in test_args:
        # Do this as a multi step replacement procedure to avoid UFL nagging
        # about too much stuff in reconstructing the expressions

        # 1) We first cut the expression to the relevant modified test_function
        # by replacing all other test functions with zero
        replacement = {ma.expr: Zero(shape=ma.expr.ufl_shape,
                                     free_indices=ma.expr.ufl_free_indices,
                                     index_dimensions=ma.expr.ufl_index_dimensions)
                       for ma in test_args}
        replacement[test_arg.expr] = test_arg.expr
        replace_expr = replace_expression(expr, replacemap=replacement)

        # 2) Propagate indexed zeros to simplify expression
        replace_expr = zero_propagation(replace_expr)

        # 3) Cut the test function itself from the expression
        #
        # This is done by replacing the test function with an
        # appropriate product of identity matrices. This way we can
        # make sure that the indices of the result will be right. This
        # is best explained by an example:
        #
        # Suppose we have the following expression:
        #
        # \sum_{i,j} a_{i,j} (\nabla v)_{i,j} + \sum_{k,l} b_{k,l} (\nable v)_{k,l}
        #
        # If we want to cut the gradient of the test function v we
        # need to make sure, that both sums have the right indices:
        #
        # \sum_{m,n} (a_{m,n} + b_{m,n}) (\nabla v)_{m,n}
        #
        # and we extract (a_{m,n} + b_{m,n}). We achieve that by the
        # following replacements:
        #
        # (\nabla v)_{i,j} -> I_{m,i} I_{n,j}
        # (\nabla v)_{k,l} -> I_{m,k} I_{n,l}
        #
        # Resulting in:
        #
        # \sum_{i,j} a_{i,j} I_{m,i} I_{n,j} + \sum_{k,l} b_{k,l} I_{m,k} I_{n,l}
        #
        # In step 4 this will collaps to: a_{m,n} + b_{m,n}
        replacement = {}
        indexmap = {}
        newi = None
        # Get all appearances of test functions with their indices
        indexed_test_args = extract_modified_arguments(replace_expr, argnumber=0, do_index=True)
        for indexed_test_arg in indexed_test_args:
            if indexed_test_arg.index:
                # If the test function is indexed, create a new multiindex of this shape
                # -> (m,n) in the example above
                if newi is None:
                    newi = indices(len(indexed_test_arg.index))
                # Replace indexed test function with a product of identities.
                identities = tuple(Indexed(Identity(2), MultiIndex((i,) + (j,)))
                                   for i, j in zip(newi, indexed_test_arg.index._indices))
                replacement.update({indexed_test_arg.expr:
                                    construct_binary_operator(identities, Product)})
                indexmap.update({i: j for i, j in zip(indexed_test_arg.index._indices, newi)})
                indexed_test_arg = analyse_modified_argument(reindexing(indexed_test_arg.expr,
                                                                        replacemap=indexmap))
            else:
                replacement.update({indexed_test_arg.expr: IntValue(1)})
        replace_expr = replace_expression(replace_expr, replacemap=replacement)

        # 4) Collapse any identity nodes that may have been introduced by replacing vectors
        replace_expr = identity_propagation(replace_expr)

        # 5) Further split according to trial function in jacobian terms
        #
        # Note: We need to split according to the FunctionView. For
        # example in Stokes with test functions v and q we have to
        # split between those.
        #
        # But: We don't want to split according to gradients since
        # this would break the input buffers of stage 3 of
        # sumfactorization
        if all_jacobian_args:
            trial_args = extract_modified_arguments(replace_expr,
                                                    argnumber=1,
                                                    do_index=False,
                                                    do_gradient=False)
            for trial_arg in trial_args:
                # 5.1) Restrict to this trial argument
                replacement = {ma.expr: Zero(shape=ma.expr.ufl_shape,
                                             free_indices=ma.expr.ufl_free_indices,
                                             index_dimensions=ma.expr.ufl_index_dimensions)
                               for ma in trial_args}
                replacement[trial_arg.expr] = trial_arg.expr
                jac_expr = replace_expression(replace_expr, replacemap=replacement)

                # 5.2) Propagate indexed zeros to simplify expression
                jac_expr = zero_propagation(jac_expr)

                # 5.3) Accumulate according to restriction
                indexed_jac_args = extract_modified_arguments(jac_expr, argnumber=1, do_index=True)
                for restriction in (Restriction.NONE, Restriction.POSITIVE, Restriction.NEGATIVE):
                    replacement = {ma.expr: Zero(shape=ma.expr.ufl_shape,
                                                 free_indices=ma.expr.ufl_free_indices,
                                                 index_dimensions=ma.expr.ufl_index_dimensions)
                                   if ma.restriction != restriction else ma.expr
                                   for ma in indexed_jac_args}

                    acc_expr = replace_expression(jac_expr, replacemap=replacement)

                    if not isinstance(jac_expr, Zero):
                        ret.append(AccumulationTerm(acc_expr, test_arg, indexmap, newi))
        else:
            if not isinstance(replace_expr, Zero):
                ret.append(AccumulationTerm(replace_expr, test_arg, indexmap, newi))

    return ret
