"""
The Jacobian Inverse Transposed for axiparallel cube grids is diagonal.
We enforce this through a post-processing transformation (instead of hacking it into UFL)
"""

from dune.perftool.ufl.transformations import ufl_transformation
from ufl.algorithms import MultiFunction
from ufl.classes import (Indexed,
                         JacobianInverse,
                         MultiIndex,
                         Product,
                         Restricted,
                         )


class DiagonalJITReplacer(MultiFunction):
    call = MultiFunction.__call__

    def __call__(self, o, delete, replace):
        self.delete = delete
        self.replace = replace

        return self.call(o)

    def expr(self, o):
        return self.reuse_if_untouched(o, *tuple(self.call(op) for op in o.ufl_operands))

    def multi_index(self, o):
        return MultiIndex(tuple(self.replace.get(i, i) for i in o))

    def index_sum(self, o):
        if o in self.delete:
            return self.call(o.ufl_operands[0])
        else:
            return self.reuse_if_untouched(o, *tuple(self.call(op) for op in o.ufl_operands))


class DiagonalJITFinder(MultiFunction):
    call = MultiFunction.__call__

    def __call__(self, o):
        self.deleted_index_sums = []
        self.replacemap = {}

        self.call(o)

        return DiagonalJITReplacer()(o, self.deleted_index_sums, self.replacemap)

    def expr(self, o):
        for op in o.ufl_operands:
            self.call(op)

    def index_sum(self, o):
        e, i = o.ufl_operands
        if isinstance(e, Product):
            p1, p2 = e.ufl_operands
            if isinstance(p1, Indexed):
                p, ii = p1.ufl_operands
                if isinstance(p, Restricted):
                    p = p.ufl_operands[0]
                if isinstance(p, JacobianInverse):
                    assert(i[0] == ii[0])
                    if ii[0] in self.replacemap:
                        self.replacemap[ii[1]] = self.replacemap[ii[0]]
                    else:
                        self.replacemap[ii[0]] = ii[1]
                    self.deleted_index_sums.append(o)
                    return

        for op in o.ufl_operands:
            self.call(op)


@ufl_transformation(name='axiparallel')
def diagonal_jacobian(e):
    """
    This transformations can generically be described as:
    \sum_k J^{(+/-)}_{k,i}x_k ==> J^{(+/-)}_{i,i}x_i
    """
    return DiagonalJITFinder()(e)
